module audio.player;
import com.cpu;
import ct.base;
import audio.audio;
import audio.filter;
import seq.sequencer;
import ui.ui;
import derelict.sdl.sdl;
import std.stdio;

enum Status { Stop, Play, Keyjam };
const int FAST_STEP = 16;
private int status;
int sec, min;
int clockCounter;
int cyclesPerFrame;
int linesPerFrame;
private int frameCallCounter, totalFrameCallCounter; // for multispeed
private int avg1,avg2,avg3,avg4; // for rastertime count
private int tickCounter;
private int fplayTickCounter, fplayRowCounter; // how many rows done since last fplay update?
int[3] muted;
bool audioinited = false;
bool dumpFrameRequested = false;
bool dumpRequested = false;
int usefp = 1, sidtype, freq = 48000, bufsize = 2048, interpolate = 1, display, badline, ntsc;
Filterparams curfp;
int curfp6581 = 0, curfp8580 = 0;


bool isPlaying() { 
	return status == Status.Play || status == Status.Keyjam;
}

bool keyjamEnabled() {
	return status == Status.Keyjam;
}

int readTick() {
	int t = tickCounter;
	tickCounter = 0;
	return t;
}

int readRowTick() {
	int t = fplayRowCounter;
	fplayRowCounter = 0;
	return t;
}


void init() {
	if(!audioinited && audio_init(bufsize, freq, ntsc ? 60 : 50, &audio_frame) < 0) {
		throw new Error("Could not init audio.");
	}
	SDL_LockAudio();
	curfp = sidtype ? FP8580[curfp8580] : FP6581[curfp6581];
	sid_init(usefp, &curfp, freq, sidtype, ntsc, interpolate, 0);
	if(!audioinited) {
		writefln("audio init: engine=%s, freq=%d, buf=%d, sid=%d, clock=%s, interpolation=%s%s%s",
				 usefp ? "resid-fp" : "resid", 
				 audio.audio.audiospec.freq, audio.audio.buffersize,
				 sidtype ? 8580 : 6581,
				 ntsc ? "ntsc" : "pal",
				 interpolate ? "on" : "off" ,
				 badline ? ", badlines=on" : "",
				 audio.audio.accurate ? ", altdelay=on" : "");
	}
	if(badline) {
		audio.audio.residdelay = 48;  // 4
	}
	else audio.audio.residdelay = 0;
	audioinited = true;
	SDL_UnlockAudio();
}

void uninit() {}

void setSidModel(int v) {
	assert(v == 0 || v == 1);
	if(sidtype == v) return;
	sidtype = v;
	init();
}

void toggleSIDModel() {
	setSidModel(sidtype ^ 1);
}

void out16b(int offs, int value) {
	song.buffer[offs] = value & 255;
	song.buffer[offs + 1] = (value >> 8) & 255;
}

private void initPlayOffset(int[3] t, int[3] s) {
	address off1 = song.offsets[Offsets.Track1] + t[0] * 2;
	address off2 = song.offsets[Offsets.Track2] + t[1] * 2;
	address off3 = song.offsets[Offsets.Track3] + t[2] * 2;
	int tpoin2 = song.offsets[Offsets.Songsets];
	int tpoin = song.offsets[Offsets.TRACKLO];
	stop();
	song.cpu.reset();
	if(song.ver >= 6) {
		song.buffer[tpoin] = off1 & 255;
		song.buffer[tpoin+1] = off2 & 255;
		song.buffer[tpoin+2] = off3 & 255;
		song.buffer[tpoin+3] = off1 >> 8;
		song.buffer[tpoin+4] = off2 >> 8;
		song.buffer[tpoin+5] = off3 >> 8;
		out16b(tpoin2, song.offsets[Offsets.Track1]);
		out16b(tpoin2+2, song.offsets[Offsets.Track2]);
		out16b(tpoin2+4, song.offsets[Offsets.Track3]);
	}
 	else {
		out16b(tpoin2, off1);
		out16b(tpoin2+2, off2);
		out16b(tpoin2+4, off3);
	}
	//SDL_LockAudio();
	cpuCall(0x1000,true);
	//SDL_UnlockAudio();
 	int seqcnt = song.offsets[Offsets.NEWSEQ];
	song.buffer[seqcnt] = s[0] * 4 + 1;
	song.buffer[seqcnt+1] = s[1] * 4 + 1;
	song.buffer[seqcnt+2] = s[2] * 4 + 1;
	song.setVoicon(muted);
}

void playNote(Element emt) {
	if(status == Status.Play) return;
	int v = seq.sequencer.activeVoiceNum;

	status = Status.Stop;
	frameCallCounter = 0;
	song.setVoicon([v == 0 ? 0 : 1, v == 1 ? 0 : 1, v == 2 ? 0 : 1]);
	muteSID(1,1,1);
	song.cpu.reset();
	song.cpu.regs.a = emt.note.value;
	song.cpu.regs.x = v; //vseq.activeVoiceNum;
	song.cpu.regs.y = emt.instr.value;
	int call = 0x1009;
	if(song.ver > 7) {
		call = song.offsets[Offsets.Subnoteplay];
	}
	cpuCall(call,true);
	status = Status.Keyjam;
}

void start(int[3] trk, int[3] seq) {
	initPlayOffset(trk,seq);
	sec = min = clockCounter = 0;
	totalFrameCallCounter = 0;
	fplayTickCounter = fplayRowCounter = 0;
	status = Status.Play;
}

void start() {
	start([0, 0, 0], [0, 0, 0]);
}

void stop() {
	status = Status.Stop;
	muteSID(1,1,1);
	frameCallCounter = 0;
	audio.audio.reset();
}

void toggleVoice(int v) {
	if(v > 2 || v < 0) return;
	muted[v] ^= 1;
	setVoicon(muted);
}

void setVoicon(int v1, int v2, int v3) {
	setVoicon([v1, v2, v3]);
}

void setVoicon(int[] m) {
	muted[0] = m[0];
	muted[1] = m[1];
	muted[2] = m[2];
	muteSID(m[0], m[1], m[2]);
	song.setVoicon(muted);
}

private void muteSID(int v1, int v2, int v3 ) {
	if(v1) song.sidbuf[4] = 0x08;
	if(v2) song.sidbuf[7 + 4] = 0x08;
	if(v3) song.sidbuf[14 + 4]= 0x08;
}

void initFP() {
	init();
	song.fppres = sidtype ? curfp8580 : curfp6581;
}

void nextFP() {
    if (usefp) {
        if (sidtype) {
            ++curfp8580;
            curfp8580 %= FP8580.length;
        } else {
            ++curfp6581;
            curfp6581 %= FP6581.length;
        }

        initFP();
    }
}


void setFP(int fp) {
    if (usefp) {
        if (sidtype) {
			curfp8580 = fp % FP8580.length;
        } else {
			curfp6581 = fp % FP6581.length;
        }
        initFP();
    }
}


void prevFP() {
    if (usefp) {
        if (sidtype) {
            --curfp8580;
            if (curfp8580 < 0) curfp8580 = FP8580.length-1;
        } else {
            --curfp6581;
            if (curfp6581 < 0) curfp6581 = FP6581.length-1;
        }

        initFP();
    }
}

void fastForward(int val) {
	int step = val * FAST_STEP;
	SDL_LockAudio();
	for(int i = 0 ; i < step; i++) {
		audio_frame();
	}
	SDL_UnlockAudio();
}

void dumpFrame() {
	if(status == Status.Play || status == Status.Keyjam)
		dumpRequested = true;
}


private void cpuCall(int pc, bool lockAudio) { cpuCall(pc, lockAudio, false); }
private void cpuCall(int pc, bool lockAudio, bool forcedump) {
	static int or1, or2, or3, or4;
	int r;
	if(lockAudio) SDL_LockAudio();
	try {
		if(dumpFrameRequested) {
			w("Frame %d(%d)", totalFrameCallCounter,frameCallCounter);
		}
		cyclesPerFrame += song.cpu.execute(pc, forcedump ? true : dumpFrameRequested);
	}
	catch(CPUException e) {
		UI.statusline.display(e.toString());
		stop();
	}
	finally {
		if(lockAudio) SDL_UnlockAudio();
	}
}

private void frameDone() {
	static int avg1, avg2, avg3, avg4;
	avg1 = avg2; avg2 = avg3; avg3 = avg4;
	avg4 = cyclesPerFrame;
	cyclesPerFrame = 0;
	linesPerFrame = (avg1 + avg2 + avg3 + avg4) / 4 / 63;

	if(dumpRequested) {
		dumpFrameRequested = true;
		dumpRequested = false;
	}
	else dumpFrameRequested = false;

	tickCounter++;
	if(++fplayTickCounter > song.playSpeed()) {
		fplayRowCounter++;
		fplayTickCounter = 0;
	}
}

void setMultiplier(int m) {
	if(m < 1 || m > 16) return;
	
	song.multiplier = m;
	audio.audio.setMultiplier(m);
}

void decMultiplier() {
	setMultiplier(song.multiplier - 1);
}

void incMultiplier() {
	setMultiplier(song.multiplier + 1);
}
	


int delayhack;
// called each frame from soundbuffer callback
extern(C) void audio_frame() {
	if(++delayhack < 0) return;
	delayhack = 0;
	switch(status) {
	case Status.Play:
		if(++clockCounter >= audio.audio.framerate) {
			clockCounter = 0;
			if(++sec > 59) {
				sec = 0; min++;
				min %= 100;
			}
		}
		cpuCall(frameCallCounter > 0 ? 0x1006 : 0x1003, false);
		break;
	case Status.Keyjam:
		address call = 0x100c;
		song.cpu.regs.x = 0;
		if(song.ver > 7) {
			call = song.offsets[Offsets.Submplayplay];
		}
		cpuCall(call,false);
		break;
	default:
		return;
	}

	frameCallCounter++;
	totalFrameCallCounter++;
	if(frameCallCounter >= audio.audio.multiplier) {
		frameCallCounter = 0;
		frameDone();
	}

	for(int i=0; i<0x19; i++) {
		sidreg[i] = song.sidbuf[i];
	}
}

