module main;
import derelict.sdl.sdl;
import com.fb;
import ui.ui;
import ui.input;
import audio.player;
import audio.filter;
import audio.audio;
import std.file;
import std.stdio;
import std.conv;
import std.string;
import std.c.stdlib;
import std.c.string;

version(linux) {
	const DIR_SEPARATOR = '/';
}

version(darwin) {
	const DIR_SEPARATOR = '/';
}

version(Win32) {
	const DIR_SEPARATOR = '\\';
}

UI mainui;

class ArgumentError : Exception {
	this(string msg) {
		super("Argument error: " ~ msg);
	}
}

void mainloop() {
	int mods, key, unicode;
	int ptick;
	bool quit = false;
	SDL_Event evt;
	while(!quit) {
		mainui.timerEvent(audio.player.readTick());
		while(SDL_PollEvent(&evt)) {
			switch(evt.type) {
			case SDL_QUIT:
				quit = true;
				break;
			case SDL_KEYDOWN:
				if(mainui.activeInput() !is null) {
					Cursor cursor = mainui.activeInput().cursor;
					if(cursor !is null) cursor.reset();
				}
				mods = evt.key.keysym.mod;
				key = evt.key.keysym.sym;
				unicode = evt.key.keysym.unicode;
				mods &= 0xffff - KMOD_NUM;
				
				if(mainui.keypress(Keyinfo(key, mods, unicode)) == EXIT)
					quit = true;

				mainui.update();
				break;
			case SDL_KEYUP:
				mods = evt.key.keysym.mod;
				key = evt.key.keysym.sym;
				unicode = evt.key.keysym.unicode;
				mods &= 0xffff - KMOD_NUM;
				mainui.keyrelease(Keyinfo(key, mods, unicode));
			case SDL_MOUSEBUTTONDOWN:
				switch(evt.button.button) {
				case 1, 3:
					int x, y;
					SDL_GetMouseState(&x, &y);
					x *= video.scalex;
					y *= video.scaley;
					int cx = (x + 4) / 8, cy = y / 14;
					mainui.clickedAt(cx, cy, evt.button.button);
					break;
				case 5:
					//rootwin.windowByCoord(cx, cy).mousewheelDown();
					mainui.keypress(Keyinfo(SDLK_DOWN, 0, 0));
					break;
				case 4:
					mainui.keypress(Keyinfo(SDLK_UP, 0, 0));
					break;
				default:
					break;
				}
				mainui.update();
				break;
			case SDL_MOUSEMOTION:
				break;
			case SDL_ACTIVEEVENT:
				break;
			case SDL_VIDEOEXPOSE:
				mainui.update();
				break;
			default:
				//writeln("Unknown SDL event ",evt.type);
				break;
			}
		}
		if(mainui.activeInput() !is null) {
			mainui.activeInput().update();
			Cursor cursor = mainui.activeInput().cursor;
			if(cursor !is null) cursor.blink();
		}
		SDL_Delay(40);
		com.fb.update();
	}
}

void printheader() {
	writefln(TITLE, " (C) 2009-12 Abaddon");
	writefln("Released under GNU GPL.");
	writefln();
	writefln("Usage: ccutter [OPTION]... [FILE]");
	writefln();
	writefln("Options:");
	writefln("  -b [value]     Set playback buffer size (def=%d)", audio.player.bufsize);
	// writefln("  -d [0|1]       Specify screen resolution (800x600/1024x768) (def=0)");
	writefln("  -e             Use slightly more accurate sid write delays");
	writefln("  -f             Start in fullscreen mode");
	writefln("  -nofp          Do not use resid-fp emulation");
	writefln("  -fpr [x]       Specify filter preset. x = 0..16 for 6581 and 0..1 for 8580");
	writefln("  -i             Disable resid interpolation (use fast mode instead)");
	writefln("  -m [0|1]       Specify SID model for reSID (6581/8580) (def=0)");
	writefln("  -n             Enable NTSC mode");
	writefln("  -r [value]     Set playback frequency (def=48000)");
	writefln("  -y             Use a YUV video overlay");
	writefln("  -ya            Keep real aspect ratio on YUV overlay (implies -y)");
	writefln();
}

int main(char[][] args) {
	int i;
	bool fs = false;
	bool yuvOverlay, keepAspect;
	string filename;
	bool fnDefined = false;

	DerelictSDL.load();
	scope(exit) SDL_Quit();

	i = 1;
	while(i < args.length) {
		switch(args[i])
		{
		case "-h", "-help", "--help", "-?":
			printheader();
			return 0;
			/+
		case "-d":
			display = toInt(args[i+1]);
			if(display > 1)
				throw new Error("Incorrect display value");
			i++;
			break;
			+/
		case "-m":
			sidtype = toInt(args[i+1]);
			if(sidtype != 0 && sidtype != 1 && sidtype != 6581 && sidtype != 8580)
				throw new ArgumentError("Incorrect SID type; specify 0 for 6581 or 1 for 8580");
			i++;
			break;
        case "-fpr":
            int fprarg = toInt(args[i+1]);
            sidtype ? (audio.player.curfp8580 = fprarg % FP8580.length) :
                        (audio.player.curfp6581 = fprarg % FP6581.length);
            i++;
			break;
		case "-i":
			audio.player.interpolate = 0;
			break;
		case "-l":
			audio.player.badline = 1;
			break;
		case "-n":
			audio.player.ntsc = 1;
			break;
		case "-r":
			audio.player.freq = toInt(args[i+1]);
			i++;
			break;
		case "-b":
			audio.player.bufsize = toInt(args[i+1]);
			i++;
			break;
		case "-f","--full":
			fs = true;
			break;
		case "-nofp":
			audio.player.usefp = 0;
			break;
		case "-e":
			audio.audio.accurate = 1;
			break;
		case "-y":
			yuvOverlay = true;
			break;
		case "-ya":
			yuvOverlay = true;
			keepAspect = true;
			break;
		default:
			if(args[i][0] == '-')
				throw new ArgumentError("Unrecognized option '" ~ args[i] ~ "'");
			if(fnDefined)
				throw new ArgumentError("Filename already defined.");
			filename = args[i];
			if(std.file.exists(filename) == 0 || std.file.isdir(filename)) {
				throw new Error("File not found!");
			}		
			fnDefined = true;
			break;
		}
		i++;
	}
	com.fb.init(fs, display, yuvOverlay, keepAspect, "CheeseCutter BETA");
	audio.player.init();
	mainui = new UI();
	if(filename) {
		string dir, fn;
		int sep = filename.rfind(DIR_SEPARATOR); 
		fn = filename[sep + 1..$];
		if(sep >= 0)
			dir = filename[0 .. sep];
		else dir = ".";
		chdir(dir);
		mainui.loadCallback(fn);
		mainui.update();
	}
	com.fb.update();
	SDL_PauseAudio(0);
	mainloop();
	audio.audio.audio_close();
	audio.player.uninit();
	return 0;   
}
