module seq.tracktable;
import seq.sequencer;
import seq.seqtable;
import com.fb;
import ui.ui;
import ui.input;
import derelict.sdl.sdl;
import std.string;
import ct.base;


class TrackVoice : SeqVoice {
	const InputTrack trackinput;
	bool displayTracklist = false;
	this(VoiceInit v) {		
		super(v);
		refreshPointer(0);
		trackinput = new InputTrack(activeRow);
		trackinput.setCoord(area.x,0);
		activeInput = trackinput;
	}

	override void trackFlush(int y) {
		trackinput.flush();
		refreshPointer(y);
	}

	override int keypress(Keyinfo key) {
		if(key.mods & KMOD_SHIFT) {
			switch(key.raw)
			{
			case SDLK_INSERT:
				trackInsert(true);
				break;
			case SDLK_DELETE:
				trackDelete(true);
				break;
			default:
				break;
			}
		}
		return trackinput.keypress(key);
	}

	override void deactivate() {
		super.deactivate();
		with(pos) {
			if(trkOffset >= getVoiceEnd()) {
				trkOffset = 0; rowCounter = seqOffset;
			}
		}
	}

	override void update() {
		super.update();
		SequenceRowData wseq, cseq;
		int h = area.y + area.h + 1;
		int y,i;
		int trkofs = pos.trkOffset;
		int rows = 0;
		int lasttrk = tracks.getListLength();
 		int counter;
		int scry = area.y + 1; 

		void printTrack() {
			if(!displayTracklist) return;
			int fgcol = 1;
			int bgcol = trkofs == pos.mark ? 6 : trkofs == tracks.wrapOffset() ? 2 : 0;
			cprint(area.x, scry, fgcol, bgcol,
				   format("%02X%02X", 
						  wseq.trk.trans, wseq.trk.no));
			if(bgcol != 0) {
				for(int i = 0; i < 14; i++) {
					screen.setbg(area.x + i, scry, bgcol);
				}
			}
		}

		trkofs -= pos.pointerOffset + seq.sequencer.anchor;
		counter = getRowcounter(trkofs);

		if(trkofs >= 0)
			wseq = getRowData(trkofs);
		y = area.y + 1;

		while(scry <= area.y + area.height) {
			if(trkofs < 0) {
				scry++;
			}
			else if(trkofs >= lasttrk+1) {
				scry++; trkofs++;
				continue;
			}
			else {
				printTrack(); 
				counter += wseq.seq.rows;
				scry++;
			}
			trkofs++;
			if(trkofs >= 0 && trkofs <= lasttrk) {
				wseq = getRowData(trkofs, 0);
			}
		}

	}
protected:
	void refreshTrack(int po) {
		refreshPointer(po);
		trackinput.refresh(activeRow);
		refreshPointer(po);
	}
	

	void trackInsert() { trackInsert(false); }
	void trackInsert(bool doInsert) {
		Track trk = activeRow.track; // FIX
		trackinput.flush();
		{
			if(!doInsert)
				tracks.expand();
			else {
				tracks.insertAt(activeRow.trkOffset);
				activeRow.track.setValue(0x80, 0);
				if(pos.trkOffset <= pos.mark)
					pos.mark++;
			}
		}
		trackinput.init(activeRow);
	}
	
	 void trackDelete() { trackDelete(false); }
	 void trackDelete(bool doDelete) {
		trackinput.flush();
		{
			if(!doDelete) {
				
				if(tracks.getListLength() == 1) {
					tracks[0].setValue(0xa0, 0);
				}
				tracks.shrink();
			}
			else {
				if(tracks.getListLength() == 1) {
					tracks[0].setValue(0xa0, 00);
				}
				else {
					tracks.deleteAt(activeRow.trkOffset);
					if(pos.trkOffset < pos.mark)
						pos.mark--;
					if(pos.mark < 0) pos.mark = 0;
				}
			}
		}
		//refresh();
		trackinput.init(activeRow);
	}

	 void trackTrans(int d) {
		trackinput.flush();
		{
			tracks.transposeAt(activeRow.trkOffset, tracks.length, d);

		}
		trackinput.init(activeRow);
	}
}

protected abstract class BaseTrackTable : VoiceTable {
protected:
	this(Rectangle a, PosinfoTable pi) { super(a, pi); }
	override int keypress(Keyinfo key) {
		if(key.mods & KMOD_CTRL) {
			switch(key.raw)
			{
			case SDLK_INSERT:
				if(key.mods & KMOD_SHIFT) {
					for(int i = 0; i < voices.length; i++) {
						auto v = cast(TrackVoice)voices[i];
						v.trackInsert(true);
					}
				} 
				else {
					(cast(TrackVoice)activeVoice).trackInsert(false);
					jump(Jump.ToEnd,true);

				}
				return OK;
			case SDLK_DELETE:
				if(key.mods & KMOD_SHIFT) {
					for(int i = 0; i < voices.length; i++) {
						auto v = cast(TrackVoice)voices[i];
						v.trackDelete((key.mods & KMOD_SHIFT) > 0);
					} 
				}
				else { 
					(cast(TrackVoice)activeVoice).trackDelete(false);
					jump(Jump.ToEnd,true);
				}
				return OK;
			case SDLK_q:
				(cast(TrackVoice)activeVoice).trackTrans(1);
				break;
			case SDLK_a:
				(cast(TrackVoice)activeVoice).trackTrans(-1);
				break;
					
			default:
				break;
			}	
		}

		super.keypress(key);

		switch(activeVoice.keypress(key))
		{
		case WRAPL:
			stepVoice(-1);
			break;
		case WRAPR:
			stepVoice(1);
			break;
		default:
			break;
		}

		return OK;

	}

	override void toSeqEnd() {
		return;
	}

	override void toSeqStart() {
		return;
	}

	override void refresh() {
		foreach(v; voices) {
			(cast(TrackVoice)v).refreshTrack(posTable.pointerOffset);
			v.refresh();
		}
	}
	
	override void jump(int jumpto, bool center) { 
	  activeVoice.trackFlush(posTable.pointerOffset); 
	  super.jump(jumpto,center); 
	}

	override void activate() {
		super.activate();
	}	

	override void deactivate() {
		super.deactivate();
		activeVoice.deactivate();
	}

	/* custom voicestepper: voice resync needed because
	 * tracks may not be aligned
	 */
	override void stepVoice(int i) {
		int nib = 3 ^ 3 - activeVoice.activeInput.nibble;
		super.stepVoice(i);
		super.step(-activeVoice.activeRow.seqOffset,0);
		activeVoice.activeInput.nibble = nib;
	}

	override void step(int st, int extra, int height) {
		activeVoice.trackFlush(posTable.pointerOffset);
		doStep(true, st);
	}
}


protected class TrackTable : BaseTrackTable {
	this(Rectangle a, PosinfoTable pi) {
		int x = 5 + com.fb.border + a.x;
		for(int v=0;v<3;v++) {
			Rectangle na = Rectangle(x, a.y, a.height, 13 + com.fb.border);
			x += 13 + com.fb.border;

			voices[v] = new TrackVoice(VoiceInit(song.tracks[v],
				na, pi.pos[v]));
		}
		super(a, pi);
	}

	override void refresh() {
		foreach(v; voices) {
			(cast(TrackVoice)v).refreshTrack(posTable.pointerOffset);
			v.refresh();
		}
	}


	override protected void doStep(bool wrapOk, int r) {
		foreach(v; voices) {
			bool wrap = wrapOk;
			//step(1,0,1);
//			if(v is activeVoice) wrap = false;
			v.scroll(r,wrap);
		}
	}

	override int keypress(Keyinfo key) {
		switch(key.raw)
		{
		case SDLK_DOWN, SDLK_PAGEDOWN:
			activeVoice.trackFlush(posTable.pointerOffset);
			if(activeVoice.atEnd()) return OK;
			int r = activeVoice.activeRow.seq.rows;
			step(r,area.height - 1,area.height);
			centerTo(tableBot);
			refresh();
			return OK;
		case SDLK_UP, SDLK_PAGEUP:
			activeVoice.trackFlush(posTable.pointerOffset);
			int t = activeVoice.activeRow.trkOffset;
			if(t == 0) t = activeVoice.getVoiceEnd();
			SequenceRowData s = activeVoice.getRowData(t - 1, 0);
			step(-s.seq.rows, 0, area.height);
			centerTo(tableBot);
			refresh();
			return OK;
		default:
			return super.keypress(key);
		}
	}

	final void displayTracklist(bool toggleOrDisable) {
		foreach(voice; voices) {
			TrackVoice tv = cast(TrackVoice)voice;
			if(!toggleOrDisable)
				tv.displayTracklist = false;
			else tv.displayTracklist ^= 1;
		}
	}
}
